--
-- Reversing Indoord Camera for FS19
-- You can look back throw the side window when you are reversing with e.g. truck and trailer.
-- @author:    	kenny456 (kenny456@seznam.cz)
-- @history:	v1.0 - 2017-09-11 - first release
--				v1.1 - 2017-09-17 - added controls for position of reverse camera
--				v1.2 - 2017-09-27 - added option to turn on right camera
--				v1.3 - 2019-02-22 - converted to FS19
--								  - added option to disable reversing camera function
--								  - added MP support for set camera position
--				v1.4 - 2020-04-09 - added function for lean to the front window
--								  - added function for lean behind the seat
--								  - added function for move driver's seat
--								  - added function for quick quick turn camera forward / backward
-- 
ReversingCamera = {};
ReversingCamera.modDirectory  = g_currentModDirectory
ReversingCamera.confDir = getUserProfileAppPath().. "modsSettings/ReversingCamera/";

function ReversingCamera.prerequisitesPresent(specializations)
	return true
end

function ReversingCamera.initSpecialization()
end

function ReversingCamera.registerOverwrittenFunctions(vehicleType)
end

function ReversingCamera.registerFunctions(vehicleType)
	SpecializationUtil.registerFunction(vehicleType, "saveToXmlReversingCamera", 						ReversingCamera.saveToXmlReversingCamera)
	SpecializationUtil.registerFunction(vehicleType, "loadFromXmlReversingCamera", 						ReversingCamera.loadFromXmlReversingCamera)
	SpecializationUtil.registerFunction(vehicleType, "activateCamera", 									ReversingCamera.activateCamera)
	SpecializationUtil.registerFunction(vehicleType, "toggleCamera", 									ReversingCamera.toggleCamera)
	SpecializationUtil.registerFunction(vehicleType, "toggleSeat", 										ReversingCamera.toggleSeat)
end

function ReversingCamera:onRegisterActionEvents(isSelected, isOnActiveVehicle)
	local spec = self.spec_ReversingCamera
	if not spec.modInitialized then
		return
	end
	
	if g_dedicatedServerInfo ~= nil then
		return
	end
	if spec.event_IDs == nil then
		spec.event_IDs = {}
	end
	if self:getIsControlled() then
		local actions_RC1 = { InputAction.RC_RESET_POSITION, InputAction.RC_TOGGLE_HELP, InputAction.RC_TOGGLE_RIGHT, InputAction.RC_CENTER_CAMERA, InputAction.RC_TOGGLE, InputAction.RC_TOGGLE_ALL, InputAction.RC_MOVE_OUT, InputAction.RC_MOVE_IN,
							InputAction.RC_MOVE_FORWARD, InputAction.RC_MOVE_BACK, InputAction.RC_LEAN, InputAction.RC_CAMERA_FORWARD, InputAction.RC_CAMERA_BACKWARD, InputAction.RC_SWITCH_VIEW }

		for _,actionName in pairs(actions_RC1) do
			local always = (actionName == InputAction.RC_MOVE_OUT or actionName == InputAction.RC_MOVE_IN or actionName == InputAction.RC_MOVE_FORWARD or actionName == InputAction.RC_MOVE_BACK) and true or false
			local _, eventID = g_inputBinding:registerActionEvent(actionName, self, ReversingCamera.actionCallback, true, true, always, spec.reverseCamera or spec.reverseCameraR)
			spec.event_IDs[actionName] = eventID
			if g_inputBinding ~= nil and g_inputBinding.events ~= nil and g_inputBinding.events[eventID] ~= nil then
				if actionName == 'something with lower priority' then
					g_inputBinding:setActionEventTextPriority(eventID, GS_PRIO_NORMAL)
				else
					g_inputBinding:setActionEventTextPriority(eventID, GS_PRIO_VERY_HIGH)
				end
				if actionName == InputAction.RC_TOGGLE_HELP then
					g_inputBinding:setActionEventTextVisibility(eventID, true)
				elseif actionName == InputAction.RC_TOGGLE or actionName == InputAction.RC_TOGGLE_ALL then
					g_inputBinding:setActionEventTextVisibility(eventID, false)
				else
					g_inputBinding:setActionEventTextVisibility(eventID, ReversingCamera.showHelp)
				end
				g_inputBinding:setActionEventActive(eventID, true)
			end
			local colliding = false
			_, colliding, _ = g_inputBinding:checkEventCollision(actionName)
			if colliding then
				print(string.format("Error: Reversing Camera got a colliding action: %s", actionName))
			end
		end
	end
end

function ReversingCamera.registerEventListeners(vehicleType)
	for _,n in pairs( { "onLoad", "onPostLoad", "saveToXMLFile", "onUpdate", "onDraw", "onRegisterActionEvents", "registerActionEventsPlayer", "onEnterVehicle", "saveToXmlReversingCamera", "loadFromXmlReversingCamera",
					"onReadStream", "onWriteStream", "toggleCamera", "toggleSeat" } ) do
		SpecializationUtil.registerEventListener(vehicleType, n, ReversingCamera)
	end
end

function ReversingCamera:onLoad(savegame)
	
	local spec = self.spec_ReversingCamera
	local xmlFile = self.xmlFile
	spec.event_IDs = {}
	spec.rcActive = true
	spec.rcCamIndex = nil
	spec.modInitialized = false
	ReversingCamera.rcActive = true
	ReversingCamera.showHelp = true
	ReversingCamera.autoBack = false
	
	if self.spec_enterable ~= nil then
		local i = 0
		local key = 'vehicle.enterable.cameras'
		while true do
			local cameraKey = key .. string.format(".camera(%d)", i)
			if not hasXMLProperty(xmlFile, cameraKey) then
				break
			end
			local inside = Utils.getNoNil(getXMLBool(xmlFile, cameraKey.."#isInside"),false);
			local rotatable = Utils.getNoNil(getXMLBool(xmlFile, cameraKey.."#rotatable"),false);
			local node = Utils.getNoNil(getXMLString(xmlFile, cameraKey.."#node"),nil);
			if inside and rotatable and node ~= nil then
				local name = getXMLString(xmlFile, cameraKey .. "#node")
				if self.i3dMappings[name] ~= nil then
					spec.internalCamera = I3DUtil.indexToObject(self.components, self.i3dMappings[name])
				else
					spec.internalCamera = I3DUtil.indexToObject(self.components, name)
				end
				spec.rcCamIndex = i + 1
				spec.modInitialized = true
				break
			end
			i = i + 1
		end
		if spec.internalCamera ~= nil then
			local p = getParent(spec.internalCamera)
			spec.cam = createTransformGroup("cam")
			link(p, spec.cam)
			link(spec.cam, spec.internalCamera)
			xt,yt,zt = getTranslation(spec.cam);
			spec.camPositionX = xt;  
			spec.camPositionY = yt; 
			spec.camPositionZ = zt;
			spec.tx=xt;
			spec.ty=yt;
			spec.tz=zt;
		else
			print('Error: ReversingCamera initialization failed for '..tostring(self:getFullName())..'! (No inside camera found )')
			return
		end
	end
	
	spec.reverseCamera = false
	spec.reverseCameraR = false
	spec.frontCamera = false
	spec.backCamera = false
	spec.revCamR = false;
	spec.transXmaxL = 0.6
	spec.transZmaxL = 0.1
	spec.rotYstartL = 4.2
	spec.transXmaxR = 0.6
	spec.transZmaxR = 0.1
	spec.rotYstartR = 2.3
	spec.seatPos = 0
	spec.frontPos = 0
	spec.backPos = 0
	spec.view = 'front'
	spec.rCamEntered = false
	
	if self.isClient and g_dedicatedServerInfo == nil then
		local configFile = ReversingCamera.confDir .. "ReversingCameraConfig.xml"
		if fileExists(configFile) then
			self:loadFromXmlReversingCamera(self:getFullName())
		else
			createFolder(getUserProfileAppPath().. "modsSettings/")
			createFolder(ReversingCamera.confDir)
			ReversingCamera.configXml = createXMLFile("ReversingCamera_XML", configFile, "ReversingCameraConfig")
			self:saveToXmlReversingCamera()
			saveXMLFile(ReversingCamera.configXml)
		end
	end
	
	local animationFile = ReversingCamera.modDirectory.."animation.xml"
	spec.animationXml = loadXMLFile("ReversingCamera_xml", animationFile);
	local i = 0
    while true do
        local key = string.format("vehicle.animations.animation(%d)", i)
        if not hasXMLProperty(spec.animationXml, key) then
            break
        end
        local animation = {}
        if self:loadAnimation(spec.animationXml, key, animation) then
			animation.parts[1].endTransOrig= {}
			animation.parts[1].endTransOrig[3] = animation.parts[1].endTrans[3]
			animation.parts[1].startTransOrig = {}
			animation.parts[1].startTransOrig[3] = animation.parts[1].startTrans[3]
            self.spec_animatedVehicle.animations[animation.name] = animation
            self.spec_animatedVehicle.animations[animation.name].parts[1].node = spec.cam
        end
        i = i + 1
    end
end

function ReversingCamera:onPostLoad(savegame)
	local spec = self.spec_ReversingCamera
	if not spec.modInitialized then
		return
	end
	
	if savegame ~= nil then
		local xmlFile = savegame.xmlFile
		local key = savegame.key ..".ReversingCamera"
		spec.transXmaxL = Utils.getNoNil(getXMLFloat(xmlFile, key.."#transXmaxL"), spec.transXmaxL)
		spec.transZmaxL = Utils.getNoNil(getXMLFloat(xmlFile, key.."#transZmaxL"), spec.transZmaxL)
		spec.transXmaxR = Utils.getNoNil(getXMLFloat(xmlFile, key.."#transXmaxR"), spec.transXmaxR)
		spec.transZmaxR = Utils.getNoNil(getXMLFloat(xmlFile, key.."#transZmaxR"), spec.transZmaxR)
		spec.revCamR = Utils.getNoNil(getXMLBool(xmlFile, key.."#revCamR"), spec.revCamR)
		spec.seatPos = Utils.getNoNil(getXMLFloat(xmlFile, key.."#seatPos"), spec.seatPos)
		spec.frontPos = Utils.getNoNil(getXMLFloat(xmlFile, key.."#frontPos"), spec.frontPos)
		spec.backPos = Utils.getNoNil(getXMLFloat(xmlFile, key.."#backPos"), spec.backPos)
		self:toggleCamera(spec.transXmaxL, spec.transZmaxL, spec.transXmaxR, spec.transZmaxR, spec.revCamR, true)
		self:toggleSeat(spec.seatPos, spec.frontPos, spec.backPos, true)
	end
end

function ReversingCamera:loadFromXmlReversingCamera(vehicle)
	local spec = self.spec_ReversingCamera
	
	local configFile = ReversingCamera.confDir .. "ReversingCameraConfig.xml"
	if self.isClient and g_dedicatedServerInfo == nil and fileExists(configFile) then
		ReversingCamera.configXml = loadXMLFile("ReversingCamera_XML", configFile)
		if getXMLBool(ReversingCamera.configXml, "ReversingCameraConfig.showHelp") ~= nil then
			ReversingCamera.showHelp = getXMLBool(ReversingCamera.configXml, "ReversingCameraConfig.showHelp")
		end;
		if getXMLBool(ReversingCamera.configXml, "ReversingCameraConfig.rcActiveGlobal") ~= nil then
			ReversingCamera.rcActive = getXMLBool(ReversingCamera.configXml, "ReversingCameraConfig.rcActiveGlobal")
		end
		if getXMLBool(ReversingCamera.configXml, "ReversingCameraConfig.autoBack") ~= nil then
			ReversingCamera.autoBack = getXMLBool(ReversingCamera.configXml, "ReversingCameraConfig.autoBack")
		end
		if vehicle ~= nil then
			local key = "ReversingCameraConfig.rcActiveVehicles"
			local i = 0
			while true do
				local name = getXMLString(ReversingCamera.configXml, "ReversingCameraConfig.rcActiveVehicles"..string.format(".vehicle(%d)", i).."#name");
				if name ~= nil then
					if name == vehicle then
						spec.rcActive = getXMLBool(ReversingCamera.configXml, "ReversingCameraConfig.rcActiveVehicles"..string.format(".vehicle(%d)", i));
						break
					end
				else
					break
				end
				i = i + 1
			end
		end
	end
end

function ReversingCamera:saveToXmlReversingCamera(vehicle, active)
	local spec = self.spec_ReversingCamera
	
	if self.isClient and ReversingCamera.configXml ~= nil and g_dedicatedServerInfo == nil then
		setXMLBool(ReversingCamera.configXml, "ReversingCameraConfig.showHelp", ReversingCamera.showHelp);
		setXMLBool(ReversingCamera.configXml, "ReversingCameraConfig.rcActiveGlobal", ReversingCamera.rcActive);
		setXMLBool(ReversingCamera.configXml, "ReversingCameraConfig.autoBack", ReversingCamera.autoBack);
		if vehicle ~= nil and active ~= nil then
			local key = "ReversingCameraConfig.rcActiveVehicles"
			local i = 0
			while true do
				local name = getXMLString(ReversingCamera.configXml, "ReversingCameraConfig.rcActiveVehicles"..string.format(".vehicle(%d)", i).."#name");
				if name ~= nil then
					if name == vehicle then
						setXMLString(ReversingCamera.configXml, "ReversingCameraConfig.rcActiveVehicles"..string.format(".vehicle(%d)", i).."#name", vehicle);
						setXMLBool(ReversingCamera.configXml, "ReversingCameraConfig.rcActiveVehicles"..string.format(".vehicle(%d)", i), active);
						break
					end
				else
					setXMLString(ReversingCamera.configXml, "ReversingCameraConfig.rcActiveVehicles"..string.format(".vehicle(%d)", i).."#name", vehicle);
					setXMLBool(ReversingCamera.configXml, "ReversingCameraConfig.rcActiveVehicles"..string.format(".vehicle(%d)", i), active);
					break
				end
				i = i + 1
				local name = key .. string.format(".vehicle(%d)", i)
				if not hasXMLProperty(ReversingCamera.configXml, name) then
					setXMLString(ReversingCamera.configXml, "ReversingCameraConfig.rcActiveVehicles"..string.format(".vehicle(%d)", i).."#name", vehicle);
					setXMLBool(ReversingCamera.configXml, "ReversingCameraConfig.rcActiveVehicles"..string.format(".vehicle(%d)", i), active);
					break
				end
			end
		end
		saveXMLFile(ReversingCamera.configXml)
	end
end

function ReversingCamera:saveToXMLFile(xmlFile, key)
	local spec = self.spec_ReversingCamera
	if not spec.modInitialized then
		return
	end
	
	setXMLFloat(xmlFile, key.."#transXmaxL", spec.transXmaxL)
	setXMLFloat(xmlFile, key.."#transZmaxL", spec.transZmaxL)
	setXMLFloat(xmlFile, key.."#transXmaxR", spec.transXmaxR)
	setXMLFloat(xmlFile, key.."#transZmaxR", spec.transZmaxR)
	setXMLBool(xmlFile, key.."#revCamR", spec.revCamR)
	setXMLFloat(xmlFile, key.."#seatPos", spec.seatPos)
	setXMLFloat(xmlFile, key.."#frontPos", spec.frontPos)
	setXMLFloat(xmlFile, key.."#backPos", spec.backPos)
end

function ReversingCamera:onUpdate(dt)
	local spec = self.spec_ReversingCamera
	if not spec.modInitialized then
		return
	end
	
	if self:getIsActive() and self:getIsControlled() then
		if self.spec_enterable.cameras[self.spec_enterable.camIndex].isInside == true and spec.internalCamera ~= nil and (spec.rcActive and ReversingCamera.rcActive) and self.spec_enterable.camIndex == spec.rcCamIndex then
			spec.reverseCamera = false
			spec.reverseCameraR = false
			spec.frontCamera = false
			spec.backCamera = false
			local math_pi_double  = math.pi*2
			local camRotY = self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY
			
			if math.floor(camRotY * 100) == math.floor(0 * 100) or math.floor(camRotY * 10) == math.floor(math.pi * 10) then
				spec.rCamEntered = false
			end
			
			if spec.rotYstartL - 0.001 > camRotY and camRotY > spec.rotYstartR + 0.001 then
				spec.frontCamera = true
			end
			if 1 > camRotY and camRotY > -1 then
				spec.backCamera = true
			end
			if math_pi_double > camRotY and camRotY > spec.rotYstartL and self:getAnimationTime('cameraForward') ~= 1 then
				spec.reverseCamera = true;
				spec.transXmax = spec.transXmaxL;
				spec.transZmax = spec.transZmaxL;
				spec.rozdilY = camRotY - spec.rotYstartL;
			end
			if spec.rotYstartR <= camRotY and camRotY <= spec.rotYstartR + 0.2 then
				spec.rCamEntered = true
			end
			if spec.rotYstartR > camRotY and camRotY > 0 and self:getAnimationTime('cameraBackward') ~= 1 and self:getAnimationTime('cameraForward') ~= 1 then
				spec.reverseCameraR = true
				spec.transXmax = spec.transXmaxR
				spec.transZmax = spec.transZmaxR
				spec.rozdilY = camRotY - spec.rotYstartR
			end
			if spec.reverseCamera or (spec.reverseCameraR and spec.revCamR and spec.rCamEntered) and not spec.input_RC_SWITCH_VIEW then
				spec.camPositionX = (spec.tx + (spec.rozdilY * 0.5 * spec.transXmax / 0.6));
				if spec.reverseCameraR == true then
					spec.camPositionZ = (spec.tz + (-1 * spec.rozdilY * 0.5 * spec.transZmax / 0.1));
				else
					spec.camPositionZ = (spec.tz + spec.seatPos + (spec.rozdilY * 0.5 * spec.transZmax / 0.1));
				end;
				if spec.reverseCameraR == true then
					if spec.camPositionX <= spec.tx - spec.transXmax then
						spec.camPositionX = spec.tx - spec.transXmax
					end
				else
					if spec.camPositionX >= spec.tx + spec.transXmax then
						spec.camPositionX = spec.tx + spec.transXmax
					end
				end;
				if spec.camPositionZ >= spec.tz + spec.transZmax then
					spec.camPositionZ = spec.tz + spec.transZmax;
				end;
				if spec.input_RC_MOVE_OUT then
					if spec.transXmax <= spec.tx + 2 then
						if spec.reverseCameraR == true then
							spec.camPositionX = spec.camPositionX - 0.01;
							spec.transXmax = spec.transXmax + 0.01;
						else
							spec.camPositionX = spec.camPositionX + 0.01;
							spec.transXmax = spec.transXmax + 0.01;
						end;
					end;
				end;
				if spec.input_RC_MOVE_IN then
					if spec.transXmax >= 0.1 then
						if spec.reverseCameraR == true then
							spec.camPositionX = spec.camPositionX + 0.01;
							spec.transXmax = spec.transXmax - 0.01;
						else
							spec.camPositionX = spec.camPositionX - 0.01;
							spec.transXmax = spec.transXmax - 0.01;
						end;
					end;
				end;
				if spec.input_RC_MOVE_FORWARD then
					if spec.reverseCameraR == true then
						if spec.transZmax <= 1.5 then
							spec.camPositionZ = spec.camPositionZ + 0.01;
							spec.transZmax = spec.transZmax + 0.01;
						end;
					else
						if spec.transZmax <= 1.5 then
							spec.camPositionZ = spec.camPositionZ + 0.01;
							spec.transZmax = spec.transZmax + 0.01
						end;
					end;
				end;
				if spec.input_RC_MOVE_BACK then
					if spec.reverseCameraR == true then
						if spec.transZmax >= 0.01 then
							spec.camPositionZ = spec.camPositionZ - 0.01
							spec.transZmax = spec.transZmax - 0.01
						end
					else
						if spec.transZmax >= 0.01 then
							spec.camPositionZ = spec.camPositionZ - 0.01
							spec.transZmax = spec.transZmax - 0.01
						end;
					end;
				end;
				if spec.reverseCameraR == true then
					spec.transXmaxR = spec.transXmax
					spec.transZmaxR = spec.transZmax
				else
					spec.transXmaxL = spec.transXmax
					spec.transZmaxL = spec.transZmax
				end
			elseif spec.frontCamera or spec.backCamera or self:getAnimationTime('cameraForward') == 1 or self:getAnimationTime('cameraBackward') == 1 then
				local animName = spec.frontCamera and 'cameraForward' or 'cameraBackward'
				local animTime = self:getAnimationTime(animName)
				local animation = self.spec_animatedVehicle.animations[animName]
				if spec.input_RC_LEAN then
					if ReversingCamera.autoBack then
						if animTime < 1 then
							self:playAnimation(animName, 1, nil, true)
						end
					else
						spec.input_RC_LEAN = false
						
						if animTime > 0 and animation.currentSpeed > 0 then
							self:playAnimation(animName, -1, nil, true)
						else
							self:playAnimation(animName, 1, nil, true)
						end
					end
				else
					if ReversingCamera.autoBack then
						if animTime > 0 then
							self:playAnimation(animName, -1, nil, true)
						end
					end
				end
				if self:getIsAnimationPlaying(animName) then
					if spec.frontCamera then
						if self.spec_enterable.cameras[self.spec_enterable.camIndex].rotX <= self.spec_enterable.cameras[self.spec_enterable.camIndex].rotMaxX then
							self.spec_enterable.cameras[self.spec_enterable.camIndex].rotX = self.spec_enterable.cameras[self.spec_enterable.camIndex].rotX + (0.015 * animation.currentSpeed * (spec.frontCamera and 1 or -1))
						else
							self.spec_enterable.cameras[self.spec_enterable.camIndex].rotX = self.spec_enterable.cameras[self.spec_enterable.camIndex].rotMaxX
						end
					else
						if self.spec_enterable.cameras[self.spec_enterable.camIndex].rotX >= self.spec_enterable.cameras[self.spec_enterable.camIndex].rotMinX then
							self.spec_enterable.cameras[self.spec_enterable.camIndex].rotX = self.spec_enterable.cameras[self.spec_enterable.camIndex].rotX + (0.015 * animation.currentSpeed * (spec.frontCamera and 1 or -1))
						else
							self.spec_enterable.cameras[self.spec_enterable.camIndex].rotX = self.spec_enterable.cameras[self.spec_enterable.camIndex].rotMinX
						end
					end
				end
				if spec.input_RC_CAMERA_FORWARD then
					spec.seatPos = spec.seatPos + 0.005
				elseif spec.input_RC_CAMERA_BACKWARD then
					spec.seatPos = spec.seatPos - 0.005
				end
				if spec.input_RC_MOVE_FORWARD then
					if spec.frontCamera then
						spec.frontPos = spec.frontPos + 0.005
					else
						spec.backPos = spec.backPos - 0.005
					end
				elseif spec.input_RC_MOVE_BACK then
					if spec.frontCamera then
						spec.frontPos = spec.frontPos - 0.005
					else
						spec.backPos = spec.backPos + 0.005
					end
				end
				if self:getAnimationTime('cameraForward') == 1 then
					spec.camPositionZ = self.spec_animatedVehicle.animations['cameraForward'].parts[1].endTransOrig[3] + spec.frontPos
					spec.camPositionY = self.spec_animatedVehicle.animations['cameraForward'].parts[1].endTrans[2]
				elseif self:getAnimationTime('cameraBackward') == 1 then
					spec.camPositionZ = self.spec_animatedVehicle.animations['cameraBackward'].parts[1].endTransOrig[3] + spec.backPos
					spec.camPositionY = self.spec_animatedVehicle.animations['cameraBackward'].parts[1].endTrans[2]
				else
					spec.camPositionX = spec.tx
					spec.camPositionY = spec.ty
					spec.camPositionZ = spec.tz + spec.seatPos
				end
			else
				spec.camPositionX = spec.tx
				spec.camPositionY = spec.ty
				spec.camPositionZ = spec.tz + spec.seatPos
			end
			if self:getAnimationTime('cameraForward') == 1 or self:getIsAnimationPlaying('cameraForward') then
				if camRotY > spec.rotYstartL - 0.15 then
					self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY = spec.rotYstartL - 0.15
				elseif camRotY < spec.rotYstartR + 0.15 then
					self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY = spec.rotYstartR + 0.15
				end
			end
			if self:getAnimationTime('cameraBackward') == 1 or self:getIsAnimationPlaying('cameraBackward') then
				if camRotY >= 0.9 then
					self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY = 0.9
				elseif camRotY <= -0.9 then
					self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY = -0.9
				end
			end
			if not spec.input_RC_SWITCH_VIEW then
				if math.floor(camRotY * 100) == math.floor(math.pi * 100 * -1) then------------------------------QuickCamera mod support
					self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY = math.pi
				
				elseif camRotY <  spec.rotYstartL - math_pi_double then
					self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY = spec.rotYstartL
				end
				if camRotY >= 6 then
					self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY = 6
				end
				if camRotY <= 0.15 and spec.revCamR and spec.rCamEntered then
					self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY = 0.15
				end
				if camRotY >= math_pi_double then
					self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY = 0
					spec.rCamEntered = false
				end
				
				if not self:getIsAnimationPlaying('cameraForward') and not self:getIsAnimationPlaying('cameraBackward') then
					setTranslation(spec.cam, spec.camPositionX, spec.camPositionY, spec.camPositionZ)
				end
			end
			
			if spec.input_RC_SWITCH_VIEW then
				spec.rCamEntered = false
				if spec.view == 'front' then
					if camRotY >= math.pi and not spec.rightSide then
						if camRotY <= math_pi_double then
							self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY = self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY + 0.1
						else
							self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY = 0
							spec.input_RC_SWITCH_VIEW = false
							spec.rightSide = false
						end
					elseif camRotY <= math.pi then
						if camRotY > 0 then
							self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY = self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY - 0.1
						else
							self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY = 0
							spec.input_RC_SWITCH_VIEW = false
							spec.rightSide = true
						end
					end
				else
					if camRotY <= 0 and not spec.rightSide then
						if camRotY >= -math.pi then
							self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY = self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY - 0.1
						else
							self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY = math.pi
							spec.input_RC_SWITCH_VIEW = false
							spec.rightSide = false
						end
					elseif camRotY >= 0 then
						if camRotY < math.pi then
							self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY = self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY + 0.1
						else
							self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY = math.pi
							spec.input_RC_SWITCH_VIEW = false
							spec.rightSide = true
						end
					end
				end
			else
				if camRotY < math.pi + math.pi / 2 and camRotY > math.pi - math.pi / 2 then
					spec.view = 'front'
					if camRotY > math.pi then
						spec.rightSide = false
					end
				else
					spec.view = 'back'
					if camRotY < 0 then
						spec.rightSide = false
					end
				end
			end
		end
	end
	if self.isClient then
			if self:getIsControlled() then
				if spec.event_IDs ~= nil and g_dedicatedServerInfo == nil then
					for actionName,eventID in pairs(spec.event_IDs) do
						g_inputBinding:setActionEventTextVisibility(eventID, ReversingCamera.showHelp)
						if actionName == InputAction.RC_TOGGLE or actionName == InputAction.RC_TOGGLE_ALL then
							g_inputBinding:setActionEventActive(eventID, true)
							g_inputBinding:setActionEventTextVisibility(eventID, false)
							
						elseif actionName == InputAction.RC_TOGGLE_RIGHT then
							g_inputBinding:setActionEventActive(eventID, spec.reverseCameraR and spec.rCamEntered and not spec.backCamera and not spec.input_RC_SWITCH_VIEW)
							g_inputBinding:setActionEventText(eventID, spec.revCamR and g_i18n:getText('RC_RIGHT_OFF') or g_i18n:getText('RC_RIGHT_ON'))
							
						elseif actionName == InputAction.RC_MOVE_OUT or actionName == InputAction.RC_MOVE_IN then
							g_inputBinding:setActionEventActive(eventID, ((spec.reverseCameraR and spec.revCamR and spec.rCamEntered) or spec.reverseCamera) and not spec.input_RC_SWITCH_VIEW)
							
						elseif actionName == InputAction.RC_MOVE_FORWARD then
							g_inputBinding:setActionEventActive(eventID, ((spec.reverseCameraR and spec.revCamR and spec.rCamEntered) or spec.reverseCamera or self:getAnimationTime('cameraForward') == 1 or self:getAnimationTime('cameraBackward') == 1) and not spec.input_RC_SWITCH_VIEW)
							g_inputBinding:setActionEventText(eventID, ((spec.reverseCameraR and spec.revCamR and spec.rCamEntered) or spec.reverseCamera) and g_i18n:getText('input_RC_MOVE_FORWARD') or g_i18n:getText('RC_VIEW_FORWARD'))
						
						elseif actionName == InputAction.RC_MOVE_BACK then
							g_inputBinding:setActionEventActive(eventID, ((spec.reverseCameraR and spec.revCamR and spec.rCamEntered) or spec.reverseCamera or self:getAnimationTime('cameraForward') == 1 or self:getAnimationTime('cameraBackward') == 1) and not spec.input_RC_SWITCH_VIEW)
							g_inputBinding:setActionEventText(eventID, ((spec.reverseCameraR and spec.revCamR and spec.rCamEntered) or spec.reverseCamera) and g_i18n:getText('input_RC_MOVE_BACK') or g_i18n:getText('RC_VIEW_BACKWARD'))
							
						elseif actionName == InputAction.RC_CAMERA_FORWARD or actionName == InputAction.RC_CAMERA_BACKWARD or actionName == InputAction.RC_SWITCH_VIEW then
	g_inputBinding:setActionEventActive(eventID, not spec.reverseCamera and (not spec.reverseCameraR or not spec.revCamR or not spec.rCamEntered) and self:getAnimationTime('cameraForward') == 0 and self:getAnimationTime('cameraBackward') == 0 and not spec.input_RC_SWITCH_VIEW)
							if actionName == InputAction.RC_SWITCH_VIEW then
								g_inputBinding:setActionEventText(eventID, spec.view == 'front' and g_i18n:getText('RC_LOOK_BACK') or g_i18n:getText('RC_LOOK_FRONT'))
							end
						
						elseif actionName == InputAction.RC_LEAN then
							g_inputBinding:setActionEventActive(eventID, (spec.frontCamera or (spec.backCamera and (not spec.revCamR or not spec.rCamEntered))) and not spec.input_RC_SWITCH_VIEW)
							g_inputBinding:setActionEventText(eventID, spec.frontCamera and (self:getAnimationTime('cameraForward') == 1 and g_i18n:getText('RC_LEAN_RETURN') or g_i18n:getText('RC_LEAN_FRONT'))
																						or (self:getAnimationTime('cameraBackward') == 1 and g_i18n:getText('RC_LEAN_RETURN') or g_i18n:getText('RC_LEAN_BACK')))
							
						elseif actionName == InputAction.RC_TOGGLE_HELP then
							g_inputBinding:setActionEventActive(eventID, spec.rcActive and ReversingCamera.rcActive)
							g_inputBinding:setActionEventTextVisibility(eventID, true)
							g_inputBinding:setActionEventText(eventID, ReversingCamera.showHelp and g_i18n:getText('RC_HELP_HIDE') or g_i18n:getText('RC_HELP_SHOW'))
						
						elseif actionName == InputAction.RC_RESET_POSITION then
g_inputBinding:setActionEventActive(eventID, ((spec.reverseCameraR and spec.revCamR and spec.rCamEntered) or spec.reverseCamera or spec.frontCamera or spec.backCamera or self:getAnimationTime('cameraForward') == 1 or self:getAnimationTime('cameraBackward') == 1) and not spec.input_RC_SWITCH_VIEW)
g_inputBinding:setActionEventText(eventID, ((spec.reverseCameraR and spec.revCamR and spec.rCamEntered) or spec.reverseCamera) and g_i18n:getText('input_RC_RESET_POSITION') or (self:getAnimationTime('cameraForward') == 1 or self:getAnimationTime('cameraBackward') == 1) and g_i18n:getText('RC_VIEW_RESET') or g_i18n:getText('RC_RESET_SEAT'))
						
						else
							g_inputBinding:setActionEventActive(eventID, spec.rcActive and ReversingCamera.rcActive)
						end
						
						if ((not spec.rcActive or not ReversingCamera.rcActive) and actionName ~= InputAction.RC_TOGGLE and actionName ~= InputAction.RC_TOGGLE_ALL) or not self:getActiveCamera().isInside or self.spec_enterable.camIndex ~= spec.rcCamIndex then
							g_inputBinding:setActionEventActive(eventID, false)
						end
					end
				end
			end
		end
end

function ReversingCamera:onDraw()
end

function ReversingCamera:actionCallback(actionName, keyStatus, arg4, arg5, arg6)
	local spec = self.spec_ReversingCamera
	if not spec.modInitialized then
		return
	end
	
	if self:getIsActive() then
		if keyStatus > 0 then
			if actionName == 'RC_RESET_POSITION' then
				if spec.reverseCamera then
					spec.transXmaxL = 0.6
					spec.transZmaxL = 0.1
					self:toggleCamera(spec.transXmaxL, spec.transZmaxL, spec.transXmaxR, spec.transZmaxR, spec.revCamR)
				elseif (spec.reverseCameraR and spec.revCamR and spec.rCamEntered) then
					spec.transXmaxR = 0.6
					spec.transZmaxR = 0.1
					self:toggleCamera(spec.transXmaxL, spec.transZmaxL, spec.transXmaxR, spec.transZmaxR, spec.revCamR)
				elseif spec.frontCamera then
					if self:getAnimationTime('cameraForward') == 0 then
						spec.seatPos = 0
						self:toggleSeat(spec.seatPos, spec.frontPos, spec.backPos, false)
					elseif self:getAnimationTime('cameraForward') == 1 then
						spec.frontPos = 0
						self:toggleSeat(spec.seatPos, spec.frontPos, spec.backPos, false)
					end
				elseif spec.backCamera then
					if self:getAnimationTime('cameraBackward') == 0 then
						spec.seatPos = 0
						self:toggleSeat(spec.seatPos, spec.frontPos, spec.backPos, false)
					elseif self:getAnimationTime('cameraBackward') == 1 then
						spec.backPos = 0
						self:toggleSeat(spec.seatPos, spec.frontPos, spec.backPos, false)
					end
				end
			elseif actionName == 'RC_TOGGLE_HELP' then
				ReversingCamera.showHelp = not ReversingCamera.showHelp
				self:saveToXmlReversingCamera()
			elseif actionName == 'RC_TOGGLE_RIGHT' then
				if spec.reverseCameraR == true then
					spec.revCamR = not spec.revCamR;
				end
				self:toggleCamera(spec.transXmaxL, spec.transZmaxL, spec.transXmaxR, spec.transZmaxR, spec.revCamR)
			elseif actionName == 'RC_CENTER_CAMERA' then
				self.spec_enterable.cameras[self.spec_enterable.camIndex].rotY = self.spec_enterable.cameras[self.spec_enterable.camIndex].origRotY
				self.spec_enterable.cameras[self.spec_enterable.camIndex].rotX = self.spec_enterable.cameras[self.spec_enterable.camIndex].origRotX
			elseif actionName == 'RC_TOGGLE' then
				if ReversingCamera.rcActive == false then
					g_currentMission:showBlinkingWarning("First activate Reversing Camera global (CTRL + ALT + R)")
				else
					local text = 'ACTIVATED'
					spec.rcActive = not spec.rcActive
					self:activateCamera(spec.rcActive, 'THIS')
					self:saveToXmlReversingCamera(self:getFullName(), spec.rcActive)
				end
			elseif actionName == 'RC_TOGGLE_ALL' then
				ReversingCamera.rcActive = not ReversingCamera.rcActive
				self:activateCamera(ReversingCamera.rcActive, 'ALL')
				self:saveToXmlReversingCamera()
			elseif actionName == 'RC_LEAN' then
				spec.input_RC_LEAN = not spec.input_RC_LEAN
			elseif actionName == 'RC_SWITCH_VIEW' then
				spec.input_RC_SWITCH_VIEW = true
			end
		end
		if actionName == 'RC_MOVE_OUT' then
			spec.input_RC_MOVE_OUT = keyStatus > 0
			if keyStatus == 0 then self:toggleCamera(spec.transXmaxL, spec.transZmaxL, spec.transXmaxR, spec.transZmaxR, spec.revCamR) end
		elseif actionName == 'RC_MOVE_IN' then
			spec.input_RC_MOVE_IN = keyStatus > 0
			if keyStatus == 0 then self:toggleCamera(spec.transXmaxL, spec.transZmaxL, spec.transXmaxR, spec.transZmaxR, spec.revCamR) end
		elseif actionName == 'RC_MOVE_FORWARD' then
			spec.input_RC_MOVE_FORWARD = keyStatus > 0
			if keyStatus == 0 then
				if self:getAnimationTime('cameraForward') == 1 or self:getAnimationTime('cameraBackward') == 1 then
					self:toggleSeat(spec.seatPos, spec.frontPos, spec.backPos, false)
				else
					self:toggleCamera(spec.transXmaxL, spec.transZmaxL, spec.transXmaxR, spec.transZmaxR, spec.revCamR)
				end
			end
		elseif actionName == 'RC_MOVE_BACK' then
			spec.input_RC_MOVE_BACK = keyStatus > 0
			if keyStatus == 0 then
				if self:getAnimationTime('cameraForward') == 1 or self:getAnimationTime('cameraBackward') == 1 then
					self:toggleSeat(spec.seatPos, spec.frontPos, spec.backPos, false)
				else
					self:toggleCamera(spec.transXmaxL, spec.transZmaxL, spec.transXmaxR, spec.transZmaxR, spec.revCamR)
				end
			end
		elseif actionName == 'RC_CAMERA_FORWARD' then
			spec.input_RC_CAMERA_FORWARD = keyStatus > 0
			if keyStatus == 0 then self:toggleSeat(spec.seatPos, spec.frontPos, spec.backPos, false) end
		elseif actionName == 'RC_CAMERA_BACKWARD' then
			spec.input_RC_CAMERA_BACKWARD = keyStatus > 0
			if keyStatus == 0 then self:toggleSeat(spec.seatPos, spec.frontPos, spec.backPos, false) end
		elseif actionName == 'RC_LEAN' then
			spec.input_RC_LEAN = keyStatus > 0
		end
	end
end

function ReversingCamera:activateCamera(status, mode)
	local spec = self.spec_ReversingCamera
	if mode == 'THIS' then
		mode = self:getFullName()
	end
	local text = 'ACTIVATED'
	if not status then
		text = 'DEACTIVATED'
		spec.reverseCamera = false
		spec.reverseCameraR = false
		self:stopAnimation('cameraForward', true)
		self:setAnimationTime('cameraForward', 0, false)
		self:stopAnimation('cameraBackward', true)
		self:setAnimationTime('cameraBackward', 0, false)
		if spec.internalCamera ~= nil then
			setTranslation(spec.cam, spec.tx, spec.ty, spec.tz)
		end
	end
	g_currentMission:showBlinkingWarning("Reversing camera function "..tostring(text).." for "..tostring(mode).." vehicle", 2000);
end

function ReversingCamera:onEnterVehicle()
	local spec = self.spec_ReversingCamera
	if not spec.modInitialized then
		return
	end
	
	if spec.rcActive == false or ReversingCamera.rcActive == false then
		spec.reverseCamera = false
		spec.reverseCameraR = false
		if spec.internalCamera ~= nil then
			setTranslation(spec.cam, spec.tx, spec.ty, spec.tz);
		end
	end
	self:loadFromXmlReversingCamera(self:getFullName())
end

function ReversingCamera:toggleCamera(transXmaxL, transZmaxL, transXmaxR, transZmaxR, revCamR, noEventSend)
	local spec = self.spec_ReversingCamera
	if not spec.modInitialized then
		return
	end

	spec.transXmaxL = transXmaxL
	spec.transZmaxL = transZmaxL
	spec.transXmaxR = transXmaxR
	spec.transZmaxR = transZmaxR
	spec.revCamR = revCamR
	
	ReversingCameraToggleEvent.sendEvent(self, transXmaxL, transZmaxL, transXmaxR, transZmaxR,  revCamR, noEventSend)
end
function ReversingCamera:toggleSeat(seatPos, frontPos, backPos, noEventSend)
	local spec = self.spec_ReversingCamera
	if not spec.modInitialized then
		return
	end

	spec.seatPos = seatPos
	spec.frontPos = frontPos
	spec.backPos = backPos
	self.spec_animatedVehicle.animations['cameraForward'].parts[1].startTrans[3] = spec.tz + seatPos
	self.spec_animatedVehicle.animations['cameraBackward'].parts[1].startTrans[3] = spec.tz + seatPos
	self.spec_animatedVehicle.animations['cameraForward'].parts[1].endTrans[3] = self.spec_animatedVehicle.animations['cameraForward'].parts[1].endTransOrig[3] + spec.frontPos
	self.spec_animatedVehicle.animations['cameraBackward'].parts[1].endTrans[3] = self.spec_animatedVehicle.animations['cameraBackward'].parts[1].endTransOrig[3] + spec.backPos
	
	ReversingCameraToggleSeatEvent.sendEvent(self, seatPos, frontPos, backPos, noEventSend)
end

function ReversingCamera:onReadStream(streamId, connection)
	local spec = self.spec_ReversingCamera
	if not spec.modInitialized then
		return
	end
	
	local transXmaxL = streamReadFloat32(streamId)
	local transZmaxL = streamReadFloat32(streamId)
	local transXmaxR = streamReadFloat32(streamId)
	local transZmaxR = streamReadFloat32(streamId)
    local revCamR = streamReadBool(streamId)
	local seatPos = streamReadFloat32(streamId)
	local frontPos = streamReadFloat32(streamId)
	local backPos = streamReadFloat32(streamId)
	if transXmaxL ~= nil or transZmaxL ~= nil or transXmaxR ~= nil or transZmaxR ~= nil or revCamR ~= nil or seatPos ~= nil or frontPos ~= nil or backPos ~= nil then
		self:toggleCamera(transXmaxL, transZmaxL, transXmaxR, transZmaxR, revCamR, true)
		self:toggleSeat(seatPos, frontPos, backPos, true)
	end
end

function ReversingCamera:onWriteStream(streamId, connection)
	local spec = self.spec_ReversingCamera
	if not spec.modInitialized then
		return
	end
	
	streamWriteFloat32(streamId, spec.transXmaxL)
	streamWriteFloat32(streamId, spec.transZmaxL)
	streamWriteFloat32(streamId, spec.transXmaxR)
	streamWriteFloat32(streamId, spec.transZmaxR)
	streamWriteBool(streamId, spec.revCamR)
	streamWriteFloat32(streamId, spec.seatPos)
	streamWriteFloat32(streamId, spec.frontPos)
	streamWriteFloat32(streamId, spec.backPos)
end

ReversingCameraToggleEvent = {};
ReversingCameraToggleEvent_mt = Class(ReversingCameraToggleEvent, Event)

InitEventClass(ReversingCameraToggleEvent, "ReversingCameraToggleEvent")

function ReversingCameraToggleEvent:emptyNew()
    local self = Event:new(ReversingCameraToggleEvent_mt)
    self.className="ReversingCameraToggleEvent"
    return self
end

function ReversingCameraToggleEvent:new(object, transXmaxL, transZmaxL, transXmaxR, transZmaxR, revCamR)
	local self = ReversingCameraToggleEvent:emptyNew()
	self.object = object
	self.transXmaxL = transXmaxL
	self.transZmaxL = transZmaxL
	self.transXmaxR = transXmaxR
	self.transZmaxR = transZmaxR
	self.revCamR = revCamR
	return self
end

function ReversingCameraToggleEvent:readStream(streamId, connection)
	self.object = NetworkUtil.readNodeObject(streamId)
	self.transXmaxL = streamReadFloat32(streamId)
	self.transZmaxL = streamReadFloat32(streamId)
	self.transXmaxR = streamReadFloat32(streamId)
	self.transZmaxR = streamReadFloat32(streamId)
    self.revCamR = streamReadBool(streamId)
    self:run(connection)
end

function ReversingCameraToggleEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.object)
	streamWriteFloat32(streamId, self.transXmaxL)
	streamWriteFloat32(streamId, self.transZmaxL)
	streamWriteFloat32(streamId, self.transXmaxR)
	streamWriteFloat32(streamId, self.transZmaxR)
	streamWriteBool(streamId, self.revCamR)
end

function ReversingCameraToggleEvent:run(connection)
	if self.object ~= nil then
		self.object:toggleCamera(self.transXmaxL, self.transZmaxL, self.transXmaxR, self.transZmaxR, self.revCamR, true)
	end
	if not connection:getIsServer() then
		g_server:broadcastEvent(ReversingCameraToggleEvent:new(self.object, self.transXmaxL, self.transZmaxL, self.transXmaxR, self.transZmaxR, self.revCamR), nil, connection, self.object)
	end
end

function ReversingCameraToggleEvent.sendEvent(vehicle, transXmaxL, transZmaxL, transXmaxR, transZmaxR, revCamR, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(ReversingCameraToggleEvent:new(vehicle, transXmaxL, transZmaxL, transXmaxR, transZmaxR,  revCamR), nil, nil, vehicle)
		else
			g_client:getServerConnection():sendEvent(ReversingCameraToggleEvent:new(vehicle, transXmaxL, transZmaxL, transXmaxR, transZmaxR, revCamR))
		end
	end
end

ReversingCameraToggleSeatEvent = {};
ReversingCameraToggleSeatEvent_mt = Class(ReversingCameraToggleSeatEvent, Event)

InitEventClass(ReversingCameraToggleSeatEvent, "ReversingCameraToggleSeatEvent")

function ReversingCameraToggleSeatEvent:emptyNew()
    local self = Event:new(ReversingCameraToggleSeatEvent_mt)
    self.className="ReversingCameraToggleSeatEvent"
    return self
end

function ReversingCameraToggleSeatEvent:new(object, seatPos, frontPos, backPos)
	local self = ReversingCameraToggleSeatEvent:emptyNew()
	self.object = object
	self.seatPos = seatPos
	self.frontPos = frontPos
	self.backPos = backPos
	return self
end

function ReversingCameraToggleSeatEvent:readStream(streamId, connection)
	self.object = NetworkUtil.readNodeObject(streamId)
	self.seatPos = streamReadFloat32(streamId)
	self.frontPos = streamReadFloat32(streamId)
	self.backPos = streamReadFloat32(streamId)
    self:run(connection)
end

function ReversingCameraToggleSeatEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.object)
	streamWriteFloat32(streamId, self.seatPos)
	streamWriteFloat32(streamId, self.frontPos)
	streamWriteFloat32(streamId, self.backPos)
end

function ReversingCameraToggleSeatEvent:run(connection)
	if self.object ~= nil then
		self.object:toggleSeat(self.seatPos, self.frontPos, self.backPos, true)
	end
	if not connection:getIsServer() then
		g_server:broadcastEvent(self, false, connection, self.object)
	end
end

function ReversingCameraToggleSeatEvent.sendEvent(vehicle, seatPos, frontPos, backPos, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(ReversingCameraToggleSeatEvent:new(vehicle, seatPos, frontPos, backPos), nil, nil, vehicle)
		else
			g_client:getServerConnection():sendEvent(ReversingCameraToggleSeatEvent:new(vehicle, seatPos, frontPos, backPos))
		end
	end
end